# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Any, Dict, Iterator, List

from swh.scheduler.interface import SchedulerInterface
from swh.scheduler.model import ListedOrigin


from ..pattern import CredentialsType, StatelessLister
from datetime import datetime, timezone

logger = logging.getLogger(__name__)

FDroidListerPage = List[Dict[str, Any]]


class FDroidLister(StatelessLister[FDroidListerPage]):
    """List origins from the FDroid."""

    LISTER_NAME = "f_droid"
    INSTANCE = "f_droid"
    VISIT_TYPE = "f_droid"
    REPO_INDEX_V2_URL = "https://f-droid.org/repo/index-v2.json"
    PACKAGE_URL_TEMPLATE = "https://f-droid.org/packages/{package_id}"

    def __init__(
        self,
        scheduler: SchedulerInterface,
        url: str = REPO_INDEX_V2_URL,
        instance: str = INSTANCE,
        credentials: CredentialsType = None,
    ):
        super().__init__(
            scheduler=scheduler,
            credentials=credentials,
            url=url,
            instance=instance,
        )

        self.session.headers.update({"Accept": "application/json"})

    def get_pages(self) -> Iterator[FDroidListerPage]:
        data = self.http_request(self.url).json()

        if "packages" not in data or not isinstance(data["packages"], dict):
            raise RuntimeError(
                f"Invalid response from {self.url}: missing 'packages' key"
            )

        packages: Dict[str, Any] = data["packages"]
        return iter([[packages]])

    def get_origins_from_page(self, page: FDroidListerPage) -> Iterator[ListedOrigin]:
        """Convert a page of FDroidLister repositories into a list of ListedOrigins"""
        assert self.lister_obj.id is not None

        for item in page:
            for id, value in item.items():
                metadata = value.get("metadata", {})
                versions = value.get("versions", {})
                last_updated = (
                    datetime.fromtimestamp(
                        metadata.get("lastUpdated", None) / 1000,
                        tz=timezone.utc,
                    )
                    if metadata.get("lastUpdated", None) is not None
                    else None
                )

                yield ListedOrigin(
                    lister_id=self.lister_obj.id,
                    visit_type=self.VISIT_TYPE,
                    url=self.PACKAGE_URL_TEMPLATE.format(
                        package_id=id,
                    ),
                    last_update=last_updated,
                    extra_loader_arguments={
                        "metadata": metadata,
                        "versions": versions,
                    },
                )
